"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const config_1 = require("./config");
const ids_json_1 = __importDefault(require("./ids.json"));
const web3_js_1 = require("@solana/web3.js");
const fs_1 = __importDefault(require("fs"));
const os_1 = __importDefault(require("os"));
const client_1 = require("./client");
const index_1 = require("./index");
const bn_js_1 = require("bn.js");
function mm() {
    return __awaiter(this, void 0, void 0, function* () {
        // load mango group and clients
        const config = new config_1.Config(ids_json_1.default);
        const cluster = (process.env.CLUSTER || 'devnet');
        const groupName = process.env.GROUP || 'devnet.2';
        const groupIds = config.getGroup(cluster, groupName);
        if (!groupIds) {
            throw new Error(`Group ${groupName} not found`);
        }
        const mangoProgramId = groupIds.mangoProgramId;
        const mangoGroupKey = groupIds.publicKey;
        const payer = new web3_js_1.Account(JSON.parse(fs_1.default.readFileSync(process.env.KEYPAIR || os_1.default.homedir() + '/.config/solana/id.json', 'utf-8')));
        console.log(`Payer: ${payer.publicKey.toBase58()}`);
        const connection = new web3_js_1.Connection(process.env.ENDPOINT_URL || config.cluster_urls[cluster], 'processed');
        const client = new client_1.MangoClient(connection, mangoProgramId);
        const mangoGroup = yield client.getMangoGroup(mangoGroupKey);
        // TODO make it fetch all mango accounts for owner and select highest value one
        const mangoAccountPk = new web3_js_1.PublicKey(// 'FG99s25HS1UKcP1jMx72Gezg6KZCC7DuKXhNW51XC1qi',
        // 'HrBxDrXQJmoz3GrN9BTb5fdQuMDQznCL6gfuXCxmLnkK',
        // 'HksXxVU7rczrKCce4RCPuhTVFp5CQWq2sPzu7Pbu9rpn',
        // 'H834axS45Yixoj4dwmKXgvXDg8VSB1aa4UV9b5UfFScy',
        // '8ejPPTu4BqDQdMr2qbkVVpxpF1bqiJr5BQsZqoJn4SFH',
        // 'ov3DM7gPp3TzY7H2xGA8zRt2rf8LtdoQUuw7MhqHbPD',
        // '3x3r1fhx6MsbJoxCt6G591ziMNcy7w1d5n8eUuUMS7ze',
        // '3PDvbfJzAtAtThsRAoto6dvhNaJzv4zc3SkQuNPRxAok',
        // '7jUGo474GASEY25FsXPS13p1xHhsuYGBrGmhyRQW3FSd',
        'A5FK3qET4FZd9jJxVepp8Hatf3d192uVxKUsXH95s7Xi');
        // while true loop with try catch inside
        // periodically check staleness
        // periodically receive fresh OB and mango account state
        // periodically
        // TODO make it be able to quote all markets
        const marketIndex = 1;
        const perpMarketConfig = config_1.getPerpMarketByIndex(groupIds, marketIndex);
        const perpMarket = yield client.getPerpMarket(perpMarketConfig.publicKey, perpMarketConfig.baseDecimals, perpMarketConfig.quoteDecimals);
        process.on('SIGINT', function () {
            console.log('Caught keyboard interrupt. Canceling orders');
            process.exit();
        });
        const interval = 10000;
        const sizePerc = 0.1;
        const charge = 0.0005;
        const leanCoeff = 0.0005;
        // eslint-disable-next-line no-constant-condition
        while (true) {
            try {
                // get fresh data
                // get orderbooks, get perp markets, caches
                // TODO load pyth oracle for most accurate prices
                const [mangoCache, mangoAccount] = yield Promise.all([
                    // perpMarket.loadBids(connection),
                    // perpMarket.loadAsks(connection),
                    mangoGroup.loadCache(connection),
                    client.getMangoAccount(mangoAccountPk, mangoGroup.dexProgramId),
                ]);
                // TODO store the prices in an array to calculate volatility
                // Model logic
                const fairValue = mangoGroup.getPrice(marketIndex, mangoCache).toNumber();
                const equity = mangoAccount
                    .computeValue(mangoGroup, mangoCache)
                    .toNumber();
                const perpAccount = mangoAccount.perpAccounts[marketIndex];
                // TODO look at event queue as well for unprocessed fills
                const basePos = perpAccount.getBasePositionUi(perpMarket);
                // TODO volatility adjustment
                const size = Math.min((equity * sizePerc) / fairValue, 200000);
                const lean = (-leanCoeff * basePos) / size;
                const bidPrice = fairValue * (1 - charge + lean);
                const askPrice = fairValue * (1 + charge + lean);
                // TODO only requote if new prices significantly different from current
                const [nativeBidPrice, nativeBidSize] = perpMarket.uiToNativePriceQuantity(bidPrice, size);
                const [nativeAskPrice, nativeAskSize] = perpMarket.uiToNativePriceQuantity(askPrice, size);
                // cancel all, requote
                const cancelAllInstr = index_1.makeCancelAllPerpOrdersInstruction(mangoProgramId, mangoGroup.publicKey, mangoAccount.publicKey, payer.publicKey, perpMarket.publicKey, perpMarket.bids, perpMarket.asks, new bn_js_1.BN(20));
                const placeBidInstr = index_1.makePlacePerpOrderInstruction(mangoProgramId, mangoGroup.publicKey, mangoAccount.publicKey, payer.publicKey, mangoCache.publicKey, perpMarket.publicKey, perpMarket.bids, perpMarket.asks, perpMarket.eventQueue, mangoAccount.getOpenOrdersKeysInBasket(), nativeBidPrice, nativeBidSize, new bn_js_1.BN(Date.now()), 'buy', 'postOnlySlide');
                const placeAskInstruction = index_1.makePlacePerpOrderInstruction(mangoProgramId, mangoGroup.publicKey, mangoAccount.publicKey, payer.publicKey, mangoCache.publicKey, perpMarket.publicKey, perpMarket.bids, perpMarket.asks, perpMarket.eventQueue, mangoAccount.getOpenOrdersKeysInBasket(), nativeAskPrice, nativeAskSize, new bn_js_1.BN(Date.now()), 'sell', 'postOnlySlide');
                const tx = new web3_js_1.Transaction();
                tx.add(cancelAllInstr);
                tx.add(placeBidInstr);
                tx.add(placeAskInstruction);
                const txid = yield client.sendTransaction(tx, payer, []);
                console.log(`quoting successful: ${txid.toString()}`);
            }
            catch (e) {
                // TODO on keyboard interrupt cancel all and exit
                // sleep for some time and retry
                console.log(e);
            }
            finally {
                console.log(`sleeping for ${interval / 1000}s`);
                yield index_1.sleep(interval);
            }
        }
    });
}
mm();
//# sourceMappingURL=mm.js.map